# Copyrigh 2016 Cisco Systems.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import collections

import mock

from neutron.tests import base

from networking_cisco.apps.saf.common import config
from networking_cisco.apps.saf.db import dfa_db_models as dbm
from networking_cisco.apps.saf.server.services.firewall.native.drivers import (
    dev_mgr)
from networking_cisco.apps.saf.server.services.firewall.native import fw_mgr


TENANT_NAME = 'TenantA'
TENANT_ID = '0000-1111-2222-5555'
FW_ID = '0000-aaaa-bbbb-ccce'
FW_NAME = 'FwA'
POLCY_ID = '0000-aaaa-bbbb-cccc'
FW_TYPE = 'TE'
ROUTER_ID = '0000-aaaa-bbbb-5555'
RULE_ID = '0000-aaaa-bbbb-cccd'
PROTOCOL = 'tcp'
SRC_IP = '1.1.1.1'
DST_IP = '2.2.2.2'
SRC_PORT = 34
DST_PORT = 43
RULE_NAME = 'RuleA'


try:
    OrderedDict = collections.OrderedDict
except AttributeError:
    import ordereddict
    OrderedDict = ordereddict.OrderedDict


class FakeClass(object):
    """Fake class. """
    @classmethod
    def imitate(cls, *others):
        for other in others:
            for name in other.__dict__:
                try:
                    setattr(cls, name, mock.Mock())
                except (TypeError, AttributeError):
                    pass
        return cls

    @classmethod
    def set_return(cls, class_name, fn_name, return_val):
        getattr(cls, fn_name).return_value = return_val


class FwMgrTest(base.BaseTestCase):
    """A test suite to exercise the FW Mgr. """

    def setUp(self):
        """Setup for the test scripts. """
        super(FwMgrTest, self).setUp()
        self._init_values()
        config = self._fill_cfg()
        self.cfg = config.CiscoDFAConfig().cfg

        fw_mgr.FwMgr.__bases__ = (FakeClass.imitate(dev_mgr.DeviceMgr,
                                                    dbm.DfaDBMixin),)
        FakeClass.set_return(dbm.DfaDBMixin, 'get_all_fw_db', dict())
        mock.patch('networking_cisco.apps.saf.server.services.firewall.'
                   'native.fabric_setup_base.FabricBase').start()
        mock.patch('networking_cisco.apps.saf.server.dfa_openstack_helper.'
                   'DfaNeutronHelper').start()
        fw_mgr.FwMgr.events = mock.Mock()
        self.fw_mgr = fw_mgr.FwMgr(self.cfg)
        self.update_fw_db_mock = mock.patch.object(
            self.fw_mgr, 'update_fw_db_final_result').start()
        self.create_fw_dev_mock = mock.patch.object(
            self.fw_mgr, 'create_fw_device').start()
        self.update_fw_dev_mock = mock.patch.object(
            self.fw_mgr, 'update_fw_db_dev_status').start()

    def _init_values(self):
        self.tenant_name = TENANT_NAME
        self.tenant_id = TENANT_ID
        self.fw_id = FW_ID
        self.fw_name = FW_NAME
        self.policy_id = POLCY_ID
        self.fw_type = FW_TYPE
        self.router_id = ROUTER_ID
        self.rule_id = RULE_ID
        self.rule_dict = {'protocol': PROTOCOL, 'name': RULE_NAME,
                          'enabled': True, 'source_ip_address': SRC_IP,
                          'destination_ip_address': DST_IP,
                          'source_port': SRC_PORT,
                          'destination_port': DST_PORT, 'action': 'allow'}

    def _fill_cfg(self):
        config.default_firewall_opts['firewall'] = {
            'device': 'phy_asa', 'fw_mgmt_ip': '1.1.1.1',
            'fw_username': 'user', 'fw_passport': 'user',
            'fw_interface_in': 'e1/1', 'fw_interface_out': 'e1/1'}
        return config

    def test_fw_mgr_init(self):
        """Wrapper for the init. """
        pass

    def _fill_rule_data(self):
        fw_rule_dict = {'firewall_rule': {'tenant_id': self.tenant_id}}
        sub_rule_dict = fw_rule_dict['firewall_rule']
        sub_rule_dict.update({
            'firewall_policy_id': self.policy_id,
            'id': self.rule_id,
            'protocol': self.rule_dict['protocol'],
            'source_ip_address': self.rule_dict['source_ip_address'],
            'destination_ip_address': self.rule_dict['destination_ip_address'],
            'source_port': self.rule_dict['source_port'],
            'destination_port': self.rule_dict['destination_port'],
            'action': self.rule_dict['action'],
            'enabled': True, 'name': RULE_NAME})
        return fw_rule_dict

    def _test_fw_rule_create(self):
        with mock.patch.object(self.fw_mgr.fabric,
                               'prepare_fabric_fw') as prep_fab:
            fw_rule_data = self._fill_rule_data()
            self.fw_mgr.fw_rule_create(fw_rule_data, None)
            parent = mock.MagicMock()
            parent.attach_mock(prep_fab, 'prepare_fabric_fw')
            parent.attach_mock(self.create_fw_dev_mock, 'create_fw_device')
            parent.attach_mock(self.update_fw_db_mock,
                               'update_fw_db_final_result')
            parent.attach_mock(self.update_fw_dev_mock,
                               'update_fw_db_dev_status')
            prep_fab.assert_not_called()
            self.update_fw_db_mock.assert_not_called()
            self.create_fw_dev_mock.assert_not_called()
            self.update_fw_dev_mock.assert_not_called()

    def test_fw_rule_create(self):
        """Test FW rule create. """
        self._test_fw_rule_create()

    def _fill_policy_data(self, rule_dict):
        fw_policy_dict = {'firewall_policy': {
            'tenant_id': self.tenant_id, 'id': self.policy_id,
            'name': 'PolicyA', 'firewall_rules': rule_dict}}
        # Fill the above with Policy rules by calling the fill rule fn
        return fw_policy_dict

    def _test_fw_policy_create(self, with_rule=False):
        with mock.patch.object(self.fw_mgr.fabric,
                               'prepare_fabric_fw') as prep_fab:
            fw_rule_data = dict()
            if with_rule:
                fw_rule_data = []
                fw_rule_data.append(self.rule_id)
                fw_policy_data = self._fill_policy_data(fw_rule_data)
                self.fw_mgr.fw_policy_create(fw_policy_data, None)
                parent = mock.MagicMock()
                parent.attach_mock(prep_fab, 'prepare_fabric_fw')
                parent.attach_mock(self.create_fw_dev_mock, 'create_fw_device')
                parent.attach_mock(self.update_fw_db_mock,
                                   'update_fw_db_final_result')
                parent.attach_mock(self.update_fw_dev_mock,
                                   'update_fw_db_dev_status')
                prep_fab.assert_not_called()
                self.update_fw_db_mock.assert_not_called()
                self.create_fw_dev_mock.assert_not_called()
                self.update_fw_dev_mock.assert_not_called()

    def test_fw_policy_create(self):
        """Test FW policy create. """
        self._test_fw_policy_create()

    def _fill_fw_data(self, rule_dict):
        return {'firewall': {'tenant_id': self.tenant_id,
                             'id': self.fw_id,
                             'firewall_policy_id': self.policy_id,
                             'admin_state_up': True, 'name': self.fw_name,
                             'firewall_rules': rule_dict,
                             'router_ids': [self.router_id]}}

    def _fill_fw_delete_data(self):
        fw_dict = {}
        fw_dict['firewall_id'] = self.fw_id
        return fw_dict

    def _prepare_result_fw_dict(self):
        return {'rules': {self.rule_id: self.rule_dict},
                'tenant_name': self.tenant_name,
                'tenant_id': self.tenant_id, 'fw_id': self.fw_id,
                'fw_name': self.fw_name,
                'firewall_policy_id': self.policy_id,
                'fw_type': self.fw_type, 'router_id': self.router_id}

    def _test_fw_create(self, with_rule=True):
        res_fw_dict = self._prepare_result_fw_dict()
        with mock.patch('networking_cisco.apps.saf.server.'
                        'dfa_openstack_helper.DfaNeutronHelper'),\
            mock.patch.object(self.fw_mgr.fabric,
                              'prepare_fabric_fw') as prep_fab,\
            mock.patch('networking_cisco.apps.saf.db.dfa_db_models.'
                       'DfaDBMixin.get_project_name',
                       return_value=self.tenant_name):
            FakeClass.set_return(dev_mgr.DeviceMgr,
                                 'is_device_virtual', False)
            self.fw_mgr.os_helper.get_rtr_name.return_value = (
                'Cisco_TenantEdge_TenantA')
            fw_rule_data = dict()
            if with_rule:
                fw_rule_data = self._fill_rule_data()
            fw_data = self._fill_fw_data(fw_rule_data)
            self._test_fw_rule_create()
            self._test_fw_policy_create(with_rule=True)
            parent = mock.MagicMock()
            parent.attach_mock(prep_fab, 'prepare_fabric_fw')
            parent.attach_mock(self.create_fw_dev_mock, 'create_fw_device')
            parent.attach_mock(self.update_fw_db_mock,
                               'update_fw_db_final_result')
            parent.attach_mock(self.update_fw_dev_mock,
                               'update_fw_db_dev_status')
            self.fw_mgr.fw_create(fw_data, None)
        expected_calls = [
            mock.call.prepare_fabric_fw(self.tenant_id, res_fw_dict, False,
                                        'FAB_CREATE_PEND'),
            mock.call.update_fw_db_final_result(res_fw_dict.get('fw_id'),
                                                'FAB_CREATE_DONE'),
            mock.call.create_fw_device(self.tenant_id,
                                       res_fw_dict.get('fw_id'), res_fw_dict),
            mock.call.update_fw_db_dev_status(res_fw_dict.get('fw_id'),
                                              'SUCCESS')]
        parent.assert_has_calls(expected_calls, any_order=True)
        # any_order should be False,in which case it didn't work TODO(padkrish)

    def test_fw_create(self):
        """Test FW create. """
        self._test_fw_create()

    def _test_fw_create_fabric_error(self, with_rule=True):
        """Internal function to test the error case for fabric create.

        The fabric module is mocked to return an error for prepare fabric.
        """
        res_fw_dict = self._prepare_result_fw_dict()
        with mock.patch('networking_cisco.apps.saf.server.'
                        'dfa_openstack_helper.DfaNeutronHelper'),\
            mock.patch('networking_cisco.apps.saf.db.dfa_db_models.'
                       'DfaDBMixin.get_project_name',
                       return_value=self.tenant_name),\
            mock.patch.object(self.fw_mgr.fabric, 'prepare_fabric_fw',
                              return_value=False) as prep_fab:
            FakeClass.set_return(dev_mgr.DeviceMgr, 'is_device_virtual', False)
            self.fw_mgr.os_helper.get_rtr_name.return_value = (
                'Cisco_TenantEdge_TenantA')
            fw_rule_data = dict()
            if with_rule:
                fw_rule_data = self._fill_rule_data()
            fw_data = self._fill_fw_data(fw_rule_data)
            self._test_fw_rule_create()
            self._test_fw_policy_create(with_rule=True)
            parent = mock.MagicMock()
            parent.attach_mock(prep_fab, 'prepare_fabric_fw')
            parent.attach_mock(self.create_fw_dev_mock, 'create_fw_device')
            parent.attach_mock(self.update_fw_db_mock,
                               'update_fw_db_final_result')
            parent.attach_mock(self.update_fw_dev_mock,
                               'update_fw_db_dev_status')
            self.fw_mgr.fw_create(fw_data, None)
        expected_calls = [mock.call.prepare_fabric_fw(self.tenant_id,
                                                      res_fw_dict, False,
                                                      'FAB_CREATE_PEND')]
        parent.assert_has_calls(expected_calls, any_order=True)
        self.update_fw_db_mock.assert_not_called()
        self.create_fw_dev_mock.assert_not_called()
        self.update_fw_dev_mock.assert_not_called()

    def test_fw_create_fabric_error(self):
        """Test FW create. """
        self._test_fw_create_fabric_error()

    def _test_fw_create_device_error(self, with_rule=True):
        """Internal function to test the error case for device setup.

        The device driver module is mocked to return an error for device
        configuration.
        """
        res_fw_dict = self._prepare_result_fw_dict()
        with mock.patch('networking_cisco.apps.saf.server.'
                        'dfa_openstack_helper.DfaNeutronHelper'),\
            mock.patch('networking_cisco.apps.saf.db.dfa_db_models.'
                       'DfaDBMixin.get_project_name',
                       return_value=self.tenant_name),\
            mock.patch.object(self.fw_mgr.fabric,
                              'prepare_fabric_fw') as prep_fab,\
            mock.patch.object(self.fw_mgr, 'create_fw_device',
                              return_value=False) as create_fw_dev:
            FakeClass.set_return(dev_mgr.DeviceMgr, 'is_device_virtual', False)
            self.fw_mgr.os_helper.get_rtr_name.return_value = (
                'Cisco_TenantEdge_TenantA')
            fw_rule_data = dict()
            if with_rule:
                fw_rule_data = self._fill_rule_data()
            fw_data = self._fill_fw_data(fw_rule_data)
            self._test_fw_rule_create()
            self._test_fw_policy_create(with_rule=True)
            parent = mock.MagicMock()
            parent.attach_mock(prep_fab, 'prepare_fabric_fw')
            parent.attach_mock(create_fw_dev, 'create_fw_device')
            parent.attach_mock(self.update_fw_db_mock,
                               'update_fw_db_final_result')
            parent.attach_mock(self.update_fw_dev_mock,
                               'update_fw_db_dev_status')
            self.fw_mgr.fw_create(fw_data, None)
        expected_calls = [
            mock.call.prepare_fabric_fw(self.tenant_id, res_fw_dict, False,
                                        'FAB_CREATE_PEND'),
            mock.call.update_fw_db_final_result(res_fw_dict.get('fw_id'),
                                                'FAB_CREATE_DONE'),
            mock.call.create_fw_device(self.tenant_id,
                                       res_fw_dict.get('fw_id'), res_fw_dict)]
        parent.assert_has_calls(expected_calls, any_order=True)
        self.update_fw_dev_mock.assert_not_called()

    def test_fw_create_device_error(self):
        """Test FW create. """
        self._test_fw_create_device_error()

    def _test_fw_delete(self, with_rule=True):
        res_fw_dict = self._prepare_result_fw_dict()
        with mock.patch('networking_cisco.apps.saf.server.'
                        'dfa_openstack_helper.DfaNeutronHelper'),\
            mock.patch.object(self.fw_mgr,
                              'delete_fw_device') as delete_fw_dev,\
            mock.patch.object(self.fw_mgr.fabric,
                              'delete_fabric_fw') as del_fab,\
            mock.patch.object(self.fw_mgr, 'delete_fw') as del_fw,\
            mock.patch('networking_cisco.apps.saf.db.dfa_db_models.'
                       'DfaDBMixin.get_project_name',
                       return_value=self.tenant_name):
            FakeClass.set_return(dev_mgr.DeviceMgr, 'is_device_virtual',
                                 False)
            self.fw_mgr.os_helper.get_rtr_name.return_value = (
                'Cisco_TenantEdge_TenantA')
            fw_rule_data = dict()
            if with_rule:
                fw_rule_data = self._fill_rule_data()
            fw_data = self._fill_fw_data(fw_rule_data)
            self._test_fw_rule_create()
            self._test_fw_policy_create(with_rule=True)
            self._test_fw_create(with_rule=True)
            parent = mock.MagicMock()
            parent.attach_mock(del_fab, 'delete_fabric_fw')
            parent.attach_mock(delete_fw_dev, 'delete_fw_device')
            parent.attach_mock(self.update_fw_db_mock,
                               'update_fw_db_final_result')
            parent.attach_mock(self.update_fw_dev_mock,
                               'update_fw_db_dev_status')
            parent.attach_mock(del_fw, 'delete_fw')
            fw_data = self._fill_fw_delete_data()
            self.fw_mgr.fw_delete(fw_data, None)
        expected_calls = [
            mock.call.update_fw_db_final_result(res_fw_dict.get('fw_id'),
                                                'FAB_DELETE_PEND'),
            mock.call.delete_fw_device(self.tenant_id,
                                       res_fw_dict.get('fw_id'), res_fw_dict),
            mock.call.update_fw_db_dev_status(res_fw_dict.get('fw_id'), ''),
            mock.call.delete_fabric_fw(self.tenant_id, res_fw_dict, False,
                                       'FAB_DELETE_PEND'),
            mock.call.update_fw_db_final_result(res_fw_dict.get('fw_id'),
                                                'FAB_DELETE_DONE'),
            mock.call.delete_fw(res_fw_dict.get('fw_id'))]
        parent.assert_has_calls(expected_calls, any_order=True)

    def test_fw_delete(self):
        """Test FW delete. """
        self._test_fw_delete()

    def _test_fw_delete_dev_error(self, with_rule=True):
        """Internal function to test the error case for device cleanup.

        The device driver module is mocked to return an error for device
        configuration cleanup.
        """
        res_fw_dict = self._prepare_result_fw_dict()
        with mock.patch('networking_cisco.apps.saf.server.'
                        'dfa_openstack_helper.DfaNeutronHelper'),\
            mock.patch.object(self.fw_mgr, 'delete_fw_device',
                              return_value=False) as delete_fw_dev,\
            mock.patch.object(self.fw_mgr.fabric,
                              'delete_fabric_fw') as del_fab,\
            mock.patch.object(self.fw_mgr, 'delete_fw') as del_fw,\
            mock.patch('networking_cisco.apps.saf.db.dfa_db_models.'
                       'DfaDBMixin.get_project_name',
                       return_value=self.tenant_name):
            FakeClass.set_return(dev_mgr.DeviceMgr, 'is_device_virtual', False)
            self.fw_mgr.os_helper.get_rtr_name.return_value = (
                'Cisco_TenantEdge_TenantA')
            fw_rule_data = dict()
            if with_rule:
                fw_rule_data = self._fill_rule_data()
            fw_data = self._fill_fw_data(fw_rule_data)
            self._test_fw_rule_create()
            self._test_fw_policy_create(with_rule=True)
            self._test_fw_create(with_rule=True)
            parent = mock.MagicMock()
            parent.attach_mock(del_fab, 'delete_fabric_fw')
            parent.attach_mock(delete_fw_dev, 'delete_fw_device')
            parent.attach_mock(self.update_fw_db_mock,
                               'update_fw_db_final_result')
            parent.attach_mock(self.update_fw_dev_mock,
                               'update_fw_db_dev_status')
            parent.attach_mock(del_fw, 'delete_fw')
            fw_data = self._fill_fw_delete_data()
            self.fw_mgr.fw_delete(fw_data, None)
        expected_calls = [
            mock.call.update_fw_db_final_result(res_fw_dict.get('fw_id'),
                                                'FAB_DELETE_PEND'),
            mock.call.delete_fw_device(self.tenant_id,
                                       res_fw_dict.get('fw_id'), res_fw_dict)]
        parent.assert_has_calls(expected_calls, any_order=True)
        del_fab.assert_not_called()
        del_fw.assert_not_called()

    def test_fw_delete_dev_error(self):
        """Test FW delete. """
        self._test_fw_delete_dev_error()

    def _test_fw_delete_fab_error(self, with_rule=True):
        """Internal function to test the error case for fabric cleanup.

        The fabric module is mocked to return an error for delete the fabric
        configuration.
        """
        res_fw_dict = self._prepare_result_fw_dict()
        with mock.patch('networking_cisco.apps.saf.server.'
                        'dfa_openstack_helper.DfaNeutronHelper'),\
            mock.patch.object(self.fw_mgr,
                              'delete_fw_device') as delete_fw_dev,\
            mock.patch.object(self.fw_mgr.fabric, 'delete_fabric_fw',
                              return_value=False) as del_fab,\
            mock.patch.object(self.fw_mgr, 'delete_fw') as del_fw,\
            mock.patch('networking_cisco.apps.saf.db.dfa_db_models.'
                       'DfaDBMixin.get_project_name',
                       return_value=self.tenant_name):
            FakeClass.set_return(dev_mgr.DeviceMgr, 'is_device_virtual', False)
            self.fw_mgr.os_helper.get_rtr_name.return_value = (
                'Cisco_TenantEdge_TenantA')
            fw_rule_data = dict()
            if with_rule:
                fw_rule_data = self._fill_rule_data()
            fw_data = self._fill_fw_data(fw_rule_data)
            self._test_fw_rule_create()
            self._test_fw_policy_create(with_rule=True)
            self._test_fw_create(with_rule=True)
            parent = mock.MagicMock()
            parent.attach_mock(del_fab, 'delete_fabric_fw')
            parent.attach_mock(delete_fw_dev, 'delete_fw_device')
            parent.attach_mock(self.update_fw_db_mock,
                               'update_fw_db_final_result')
            parent.attach_mock(self.update_fw_dev_mock,
                               'update_fw_db_dev_status')
            parent.attach_mock(del_fw, 'delete_fw')
            fw_data = self._fill_fw_delete_data()
            self.fw_mgr.fw_delete(fw_data, None)
        expected_calls = [
            mock.call.update_fw_db_final_result(res_fw_dict.get('fw_id'),
                                                'FAB_DELETE_PEND'),
            mock.call.delete_fw_device(self.tenant_id,
                                       res_fw_dict.get('fw_id'), res_fw_dict),
            mock.call.update_fw_db_dev_status(res_fw_dict.get('fw_id'), ''),
            mock.call.delete_fabric_fw(self.tenant_id, res_fw_dict, False,
                                       'FAB_DELETE_PEND')]
        parent.assert_has_calls(expected_calls, any_order=True)
        del_fw.assert_not_called()

    def test_fw_delete_fab_error(self):
        """Test FW delete. """
        self._test_fw_delete_fab_error()



# Copyright 2016 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import time

from networking_cisco.apps.saf.common import config
from networking_cisco.apps.saf.common import dfa_logger as logging
from networking_cisco.apps.saf.server import dfa_openstack_helper as OsHelper
from networking_cisco.apps.saf.server.services.firewall.native import (
    fabric_setup_base as FP)
from networking_cisco.apps.saf.server.services.firewall.native import (
    fw_constants as fw_const)
from networking_cisco.apps.saf.server.services.firewall.native.drivers import (
    base)


LOG = logging.getLogger(__name__)
Q_PRIORITY = 30 + 4


class NativeFirewall(base.BaseDriver, FP.FabricApi):

    """Native Firewall Driver. """

    def __init__(self):
        """Class init. """
        LOG.debug("Initializing Native Firewall")
        super(NativeFirewall, self).__init__()
        self.tenant_dict = {}
        self.os_helper = OsHelper.DfaNeutronHelper()
        self.cfg = config.CiscoDFAConfig().cfg
        self.mgmt_ip_addr = None
        self.dcnm_obj = None
        self.que_obj = None

    def initialize(self, cfg_dict):
        """Initialization routine. """
        LOG.debug("Initialize for NativeFirewall")
        self.mgmt_ip_addr = cfg_dict.get('mgmt_ip_addr')

    def populate_event_que(self, que_obj):
        """Populate the event queue object. """
        LOG.debug("Pop Event for NativeFirewall")
        self.que_obj = que_obj

    def populate_dcnm_obj(self, dcnm_obj):
        """Populate the DCNM object. """
        LOG.debug("Pop DCNM for NativeFirewall")
        self.dcnm_obj = dcnm_obj

    def is_device_virtual(self):
        """Returns if device is virtual. """
        return True

    def get_name(self):
        """Returns the name of the FW appliance. """
        # Put it in a constant fixme(padkrish)
        return 'native'

    def get_max_quota(self):
        """Returns the number of Firewall instances.

        Returns the maximum number of Firewall instance that a single Firewall
        can support.
        """
        # TODO(padkrish) Return the right value
        return 50

    def attach_intf_router(self, tenant_id, tenant_name, router_id):
        """Routine to attach the interface to the router. """
        in_sub = self.get_in_subnet_id(tenant_id)
        out_sub = self.get_out_subnet_id(tenant_id)
        # Modify Hard coded Name fixme
        subnet_lst = set()
        subnet_lst.add(in_sub)
        subnet_lst.add(out_sub)
        ret = self.os_helper.add_intf_router(router_id, tenant_id, subnet_lst)
        return ret, in_sub, out_sub

    def get_router_id(self, tenant_id, tenant_name):
        """Retrieve the router ID. """
        router_id = None
        if tenant_id in self.tenant_dict:
            router_id = self.tenant_dict.get(tenant_id).get('router_id')
        if not router_id:
            router_list = self.os_helper.get_rtr_by_name(
                'FW_RTR_' + tenant_name)
            if len(router_list) > 0:
                router_id = router_list[0].get('id')
        return router_id

    def delete_intf_router(self, tenant_id, tenant_name, router_id):
        """Routine to delete the router. """
        in_sub = self.get_in_subnet_id(tenant_id)
        out_sub = self.get_out_subnet_id(tenant_id)
        subnet_lst = set()
        subnet_lst.add(in_sub)
        subnet_lst.add(out_sub)
        router_id = self.get_router_id(tenant_id, tenant_name)
        if router_id:
            ret = self.os_helper.delete_intf_router(tenant_name, tenant_id,
                                                    router_id, subnet_lst)
            if not ret:
                LOG.error("Failed to delete router intf id %(rtr)s, "
                          "tenant %(tenant)s",
                          {'rtr': router_id, 'tenant': tenant_id})
            return ret
        LOG.error("Invalid router ID, can't delete interface from "
                  "router")

    def prepare_router_vm_msg(self, tenant_id, tenant_name, router_id, net_id,
                              subnet_id, seg, status):
        """Prepare the message to be sent to Event queue for VDP trigger.

        This is actually called for a subnet add to a router. This function
        prepares a VM's VNIC create/delete message.
        """
        max_get_router_info_retry = True
        attempt = 0
        while max_get_router_info_retry:
            port_data = self.os_helper.get_router_port_subnet(subnet_id)
            if port_data is None:
                LOG.error("Unable to get router port data")
                return None
            if port_data.get('binding:host_id') == '':
                time.sleep(3)
                attempt += 1
                if attempt > 3:
                    max_get_router_info_retry = False
                    LOG.error("Unable to get router binding host data, "
                              "Max attempts reached")
            else:
                max_get_router_info_retry = False
        if status is 'up':
            event_type = 'service.vnic.create'
        else:
            event_type = 'service.vnic.delete'
        vnic_data = {'status': status, 'mac': port_data.get('mac_address'),
                     'segid': seg, 'host': port_data.get('binding:host_id')}
        if vnic_data['host'] == '':
            LOG.error("Null host for seg %(seg)s subnet %(subnet)s",
                      {'seg': seg, 'subnet': subnet_id})
            if self.tenant_dict.get(tenant_id).get('host') is None:
                LOG.error("Null host for tenant %(tenant)s seg %(seg)s "
                          "subnet %(subnet)s",
                          {'tenant': tenant_id, 'seg': seg,
                           'subnet': subnet_id})
                return None
            else:
                vnic_data['host'] = self.tenant_dict.get(tenant_id).get('host')
        else:
            self.tenant_dict[tenant_id]['host'] = vnic_data['host']
        vm_ip = port_data.get('fixed_ips')[0].get('ip_address')
        vnic_data.update({'port_id': port_data.get('id'), 'network_id': net_id,
                          'vm_name': 'FW_SRVC_RTR_' + tenant_name,
                          'vm_ip': vm_ip, 'vm_uuid': router_id, 'gw_mac': None,
                          'fwd_mod': 'anycast_gateway'})
        payload = {'service': vnic_data}
        data = (event_type, payload)
        return data

    def send_router_port_msg(self, tenant_id, tenant_name, router_id, net_id,
                             subnet_id, seg, status):
        """Sends the router port message to the queue. """
        data = self.prepare_router_vm_msg(tenant_id, tenant_name, router_id,
                                          net_id, subnet_id, seg, status)
        if data is None:
            return False
        timestamp = time.ctime()
        pri = Q_PRIORITY
        LOG.info("Sending native FW data into queue %(data)s",
                 {'data': data})
        self.que_obj.put((pri, timestamp, data))
        return True

    def create_tenant_dict(self, tenant_id, router_id=None):
        """Tenant dict creation. """
        self.tenant_dict[tenant_id] = {'host': None, 'router_id': router_id}

    def send_in_router_port_msg(self, tenant_id, arg_dict, status):
        "Call routine to send vNic create notification for 'in' interface. """
        in_net = self.get_in_net_id(tenant_id)
        ret = self.send_router_port_msg(tenant_id,
                                        arg_dict.get('tenant_name') + '_in',
                                        arg_dict.get('router_id'), in_net,
                                        arg_dict.get('in_sub'),
                                        arg_dict.get('in_seg'), status)
        if not ret:
            LOG.error("Sending router port message failed for in network "
                      "tenant %(tenant)s subnet %(seg)s",
                      {'tenant': tenant_id, 'seg': arg_dict.get('in_seg')})
            if status == 'up':
                self.delete_intf_router(tenant_id, arg_dict.get('tenant_name'),
                                        arg_dict.get('router_id'))
            return False
        return True

    def send_out_router_port_msg(self, tenant_id, arg_dict, status):
        "Call routine to send vNic create notification for 'out' interface. """
        out_net = self.get_out_net_id(tenant_id)
        router_id = arg_dict.get('router_id')
        in_seg = arg_dict.get('in_seg')
        out_seg = arg_dict.get('out_seg')
        in_sub = arg_dict.get('in_sub')
        out_sub = arg_dict.get('out_sub')
        ret = self.send_router_port_msg(tenant_id,
                                        arg_dict.get('tenant_name') + '_out',
                                        router_id, out_net, out_sub, out_seg,
                                        status)
        if not ret:
            LOG.error("Sending router port message failed for out network "
                      "tenant %(tenant)s subnet %(seg)s",
                      {'tenant': tenant_id, 'seg': out_seg})
            in_net = self.get_in_net_id(tenant_id)
            if status == 'up':
                ret = self.send_router_port_msg(
                    tenant_id, arg_dict.get('tenant_name') + '_in', router_id,
                    in_net, in_sub, in_seg, 'down')
                if not ret:
                    LOG.error("Error case, Sending router port message "
                              "failed for in network tenant %(tenant)s "
                              "subnet %(seg)s",
                              {'tenant': tenant_id, 'seg': in_seg})
                self.delete_intf_router(tenant_id, arg_dict.get('tenant_name'),
                                        router_id)
            return False
        return True

    def program_next_hop(self, tenant_id, arg_dict):
        "Program the next hop for all host subnets to the 'in' gateway. """
        if arg_dict.get('in_gw') != 0:
            ret = self.os_helper.program_rtr_all_nwk_next_hop(
                tenant_id, arg_dict.get('router_id'), arg_dict.get('in_gw'),
                arg_dict.get('excl_list'))
            if not ret:
                LOG.error("Unable to program default router next hop %s",
                          arg_dict.get('router_id'))
                self.delete_intf_router(tenant_id, arg_dict.get('tenant_name'),
                                        arg_dict.get('router_id'))
                return False
        return True

    def program_default_gw(self, tenant_id, arg_dict):
        "Program the default gateway to the 'out' interface. """
        ret = False
        attempt = 0
        if arg_dict.get('out_gw') != 0:
            while not ret and attempt <= 3:
                time.sleep(5)
                ret = self.os_helper.program_rtr_default_gw(
                    tenant_id, arg_dict.get('router_id'),
                    arg_dict.get('out_gw'))
                attempt += 1
        if not ret:
            LOG.error("Unable to program default GW in router %s",
                      arg_dict.get('router_id'))
            self.delete_intf_router(tenant_id, arg_dict.get('tenant_name'),
                                    arg_dict.get('router_id'))
            return False
        return True

    def update_dcnm_partition_static_route(self, tenant_id, arg_dict):
        """Add static route in DCNM's partition.

        This gets pushed to the relevant leaf switches.
        """
        ip_list = self.os_helper.get_subnet_nwk_excl(tenant_id,
                                                     arg_dict.get('excl_list'))
        srvc_node_ip = self.get_out_srvc_node_ip_addr(tenant_id)
        ret = self.dcnm_obj.update_partition_static_route(
            arg_dict.get('tenant_name'), fw_const.SERV_PART_NAME, ip_list,
            vrf_prof=self.cfg.firewall.fw_service_part_vrf_profile,
            service_node_ip=srvc_node_ip)
        if not ret:
            LOG.error("Unable to update DCNM ext profile with static "
                      "route %s", arg_dict.get('router_id'))
            self.delete_intf_router(tenant_id, arg_dict.get('tenant_name'),
                                    arg_dict.get('router_id'))
            return False
        return True

    def _create_arg_dict(self, tenant_id, data, in_sub, out_sub):
        """Create the argument dictionary. """
        in_seg, in_vlan = self.get_in_seg_vlan(tenant_id)
        out_seg, out_vlan = self.get_out_seg_vlan(tenant_id)
        in_ip_dict = self.get_in_ip_addr(tenant_id)
        out_ip_dict = self.get_out_ip_addr(tenant_id)
        excl_list = [in_ip_dict.get('subnet'), out_ip_dict.get('subnet')]

        arg_dict = {'tenant_id': tenant_id,
                    'tenant_name': data.get('tenant_name'),
                    'in_seg': in_seg, 'in_vlan': in_vlan,
                    'out_seg': out_seg, 'out_vlan': out_vlan,
                    'router_id': data.get('router_id'),
                    'in_sub': in_sub, 'out_sub': out_sub,
                    'in_gw': in_ip_dict.get('gateway'),
                    'out_gw': out_ip_dict.get('gateway'),
                    'excl_list': excl_list}
        return arg_dict

    def _create_fw(self, tenant_id, data):
        """Internal routine that gets called when a FW is created. """
        LOG.debug("In creating Native FW data is %s", data)
        # TODO(padkrish):
        # Check if router is already added and only then add, needed for
        # restart cases since native doesn't have a special DB
        ret, in_sub, out_sub = self.attach_intf_router(tenant_id,
                                                       data.get('tenant_name'),
                                                       data.get('router_id'))
        if not ret:
            LOG.error("Native FW: Attach intf router failed for tenant "
                      "%s", tenant_id)
            return False

        self.create_tenant_dict(tenant_id, data.get('router_id'))
        arg_dict = self._create_arg_dict(tenant_id, data, in_sub, out_sub)

        # Program DCNM to update profile's static IP address on OUT part
        ret = self.update_dcnm_partition_static_route(tenant_id, arg_dict)
        if not ret:
            return False
        # Program the default GW in router namespace
        ret = self.program_default_gw(tenant_id, arg_dict)
        if not ret:
            return False

        # Program router namespace to have all tenant networks to be routed
        # to IN service network
        ret = self.program_next_hop(tenant_id, arg_dict)
        if not ret:
            return False

        # Send message for router port auto config for in service nwk
        ret = self.send_in_router_port_msg(tenant_id, arg_dict, 'up')
        if not ret:
            return False

        # Send message for router port auto config for out service nwk
        return self.send_out_router_port_msg(tenant_id, arg_dict, 'up')

    def create_fw(self, tenant_id, data):
        """Top level routine called when a FW is created. """
        try:
            return self._create_fw(tenant_id, data)
        except Exception as exc:
            LOG.error("Failed to create FW for device native, tenant "
                      "%(tenant)s data %(data)s Exc %(exc)s",
                      {'tenant': tenant_id, 'data': data, 'exc': exc})
            return False

    def _delete_fw(self, tenant_id, data):
        """Internal routine called when a FW is deleted. """
        LOG.debug("In Delete fw data is %s", data)
        in_sub = self.get_in_subnet_id(tenant_id)
        out_sub = self.get_out_subnet_id(tenant_id)
        arg_dict = self._create_arg_dict(tenant_id, data, in_sub, out_sub)

        if arg_dict.get('router_id') is None:
            LOG.error("Router ID unknown for tenant %s", tenant_id)
            return False

        if tenant_id not in self.tenant_dict:
            self.create_tenant_dict(tenant_id, arg_dict.get('router_id'))
        ret = self.send_in_router_port_msg(tenant_id, arg_dict, 'down')
        if not ret:
            return False
        ret = self.send_out_router_port_msg(tenant_id, arg_dict, 'down')
        if not ret:
            return False
        # Usually sending message to queue doesn't fail!!!

        router_ret = self.delete_intf_router(tenant_id,
                                             arg_dict.get('tenant_name'),
                                             arg_dict.get('router_id'))
        if not router_ret:
            LOG.error("Unable to delete router for tenant %s, error case",
                      tenant_id)
            return router_ret
        del self.tenant_dict[tenant_id]
        return router_ret

    def delete_fw(self, tenant_id, data):
        """Top level routine called when a FW is deleted. """
        try:
            ret = self._delete_fw(tenant_id, data)
            return ret
        except Exception as exc:
            LOG.error("Failed to delete FW for device native, tenant "
                      "%(tenant)s data %(data)s Exc %(exc)s",
                      {'tenant': tenant_id, 'data': data, 'exc': exc})
            return False

    def modify_fw(self, tenant_id, data):
        """Modify Firewall attributes.

        Routine called when Firewall attributes gets modified. Nothing to be
        done for native FW.
        """
        LOG.debug("In Modify fw data is %s", data)

    def _program_dcnm_static_route(self, tenant_id, tenant_name):
        """Program DCNM Static Route. """
        in_ip_dict = self.get_in_ip_addr(tenant_id)
        in_gw = in_ip_dict.get('gateway')
        in_ip = in_ip_dict.get('subnet')
        if in_gw is None:
            LOG.error("No FW service GW present")
            return False
        out_ip_dict = self.get_out_ip_addr(tenant_id)
        out_ip = out_ip_dict.get('subnet')

        # Program DCNM to update profile's static IP address on OUT part
        excl_list = []
        excl_list.append(in_ip)
        excl_list.append(out_ip)
        subnet_lst = self.os_helper.get_subnet_nwk_excl(tenant_id, excl_list,
                                                        excl_part=True)
        # This count is for telling DCNM to insert the static route in a
        # particular position. Total networks created - exclusive list as
        # above - the network that just got created.
        srvc_node_ip = self.get_out_srvc_node_ip_addr(tenant_id)
        ret = self.dcnm_obj.update_partition_static_route(
            tenant_name, fw_const.SERV_PART_NAME, subnet_lst,
            vrf_prof=self.cfg.firewall.fw_service_part_vrf_profile,
            service_node_ip=srvc_node_ip)
        if not ret:
            LOG.error("Unable to update DCNM ext profile with static "
                      "route")
            return False
        return True

    def network_create_notif(self, tenant_id, tenant_name, cidr):
        """Tenant Network create Notification.

        Restart is not supported currently for this. fixme(padkrish).
        """
        router_id = self.get_router_id(tenant_id, tenant_name)
        if not router_id:
            LOG.error("Rout ID not present for tenant")
            return False
        ret = self._program_dcnm_static_route(tenant_id, tenant_name)
        if not ret:
            LOG.error("Program DCNM with static routes failed "
                      "for router %s", router_id)
            return False

        # Program router namespace to have this network to be routed
        # to IN service network
        in_ip_dict = self.get_in_ip_addr(tenant_id)
        in_gw = in_ip_dict.get('gateway')
        if in_gw is None:
            LOG.error("No FW service GW present")
            return False
        ret = self.os_helper.program_rtr_nwk_next_hop(router_id, in_gw, cidr)
        if not ret:
            LOG.error("Unable to program default router next hop %s",
                      router_id)
            return False
        return True

    def network_delete_notif(self, tenant_id, tenant_name, network_id):
        """Tenant Network delete Notification.

        Restart is not supported currently for this. fixme(padkrish).
        """
        router_id = self.get_router_id(tenant_id, tenant_name)
        if router_id is None:
            LOG.error("Rout ID not present for tenant")
            return False
        ret = self._program_dcnm_static_route(tenant_id, tenant_name)
        if not ret:
            LOG.error("Program DCNM with static routes failed for "
                      "router %s", router_id)
            return False

        # Program router namespace to have this network to be routed
        # to IN service network
        in_ip_dict = self.get_in_ip_addr(tenant_id)
        in_gw = in_ip_dict.get('gateway')
        in_ip = in_ip_dict.get('subnet')
        if in_gw is None:
            LOG.error("No FW service GW present")
            return False
        out_ip_dict = self.get_out_ip_addr(tenant_id)
        out_ip = out_ip_dict.get('subnet')
        excl_list = []
        excl_list.append(in_ip)
        excl_list.append(out_ip)
        subnet_lst = self.os_helper.get_subnet_nwk_excl(tenant_id, excl_list,
                                                        excl_part=True)
        ret = self.os_helper.remove_rtr_nwk_next_hop(router_id, in_gw,
                                                     subnet_lst, excl_list)
        if not ret:
            LOG.error("Unable to program default router next hop %s",
                      router_id)
            return False
        return True



# Copyright 2015 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Neutron routers in Cisco devices

Revision ID: 53f08de0523f
Revises: 2921fe565328
Create Date: 2015-09-28 09:10:46.191557

"""

# revision identifiers, used by Alembic.
revision = '53f08de0523f'
down_revision = 'ff1d905b4db'
depends_on = ('2921fe565328')

from alembic import op
import sqlalchemy as sa
from sqlalchemy.engine import reflection

from neutron.db import migration


def upgrade():
    op.create_table('cisco_router_types',
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=False),
        sa.Column('description', sa.String(length=255), nullable=True),
        sa.Column('template_id', sa.String(length=36), nullable=True),
        sa.Column('ha_enabled_by_default', sa.Boolean(), nullable=False,
                  server_default=sa.sql.false()),
        sa.Column('shared', sa.Boolean(), nullable=False,
                  server_default=sa.sql.true()),
        sa.Column('slot_need', sa.Integer(), autoincrement=False,
                  nullable=True),
        sa.Column('scheduler', sa.String(length=255), nullable=False),
        sa.Column('driver', sa.String(length=255), nullable=False),
        sa.Column('cfg_agent_service_helper', sa.String(length=255),
                  nullable=False),
        sa.Column('cfg_agent_driver', sa.String(length=255), nullable=False),
        sa.ForeignKeyConstraint(['template_id'],
                                ['cisco_hosting_device_templates.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('id')
    )
    if migration.schema_has_table('cisco_router_mappings'):
        op.add_column('cisco_router_mappings',
                      sa.Column('role', sa.String(255), nullable=True))
        op.add_column('cisco_router_mappings',
                      sa.Column('router_type_id', sa.String(length=36),
                                nullable=False))

        inspector = reflection.Inspector.from_engine(op.get_bind())
        foreign_keys = inspector.get_foreign_keys('cisco_router_mappings')
        migration.remove_foreign_keys('cisco_router_mappings', foreign_keys)
        primary_key = inspector.get_pk_constraint('cisco_router_mappings')
        op.drop_constraint(constraint_name=primary_key['name'],
                           table_name='cisco_router_mappings',
                           type_='primary')

        op.create_foreign_key('cisco_router_mappings_ibfk_1',
                              source_table='cisco_router_mappings',
                              referent_table='cisco_hosting_devices',
                              local_cols=['hosting_device_id'],
                              remote_cols=['id'],
                              ondelete='SET NULL'),
        op.create_foreign_key('cisco_router_mappings_ibfk_2',
                              source_table='cisco_router_mappings',
                              referent_table='routers',
                              local_cols=['router_id'],
                              remote_cols=['id'],
                              ondelete='CASCADE')
        op.create_foreign_key('cisco_router_mappings_ibfk_3',
                              source_table='cisco_router_mappings',
                              referent_table='cisco_router_types',
                              local_cols=['router_type_id'],
                              remote_cols=['id'])
        op.create_primary_key(
            constraint_name='pk_cisco_router_mappings',
            table_name='cisco_router_mappings',
            columns=['router_id', 'router_type_id'])
        op.add_column('cisco_router_mappings',
                      sa.Column('inflated_slot_need', sa.Integer(),
                                autoincrement=False, nullable=True,
                                server_default='0'))
        op.add_column('cisco_router_mappings',
                      sa.Column('share_hosting_device', sa.Boolean(),
                                nullable=False, server_default=sa.sql.true()))
        op.create_index(op.f('ix_cisco_router_types_tenant_id'),
                        'cisco_router_types', ['tenant_id'], unique=False)



# Copyright 2015 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


import time

from oslo_config import cfg
import oslo_messaging as messaging

from networking_cisco.apps.saf.common import dfa_logger as logging

LOG = logging.getLogger(__name__)


# RPC exceptions
RPCException = messaging.MessagingException
RemoteError = messaging.RemoteError
MessagingTimeout = messaging.MessagingTimeout


class DfaRpcClient(object):

    """RPC Client class for DFA enabler."""

    def __init__(self, transport_url, topic, exchange=None, fanout=False):
        super(DfaRpcClient, self).__init__()
        transport = messaging.get_transport(cfg.CONF, url=transport_url)
        target = messaging.Target(exchange=exchange,
                                  topic=topic, fanout=fanout)
        self._client = messaging.RPCClient(transport, target)

    def make_msg(self, method, context, **kwargs):
        return {'method': method,
                'context': context,
                'args': kwargs}

    def call(self, msg):
        return self._rpc_call(msg)

    def cast(self, msg):
        return self._rpc_cast(msg)

    def _rpc_call(self, msg):
        return self._client.call(msg['context'], msg['method'], **msg['args'])

    def _rpc_cast(self, msg):
        return self._client.cast(msg['context'], msg['method'], **msg['args'])


class DfaRpcServer(object):

    """RPC server class for DFA enabler."""

    def __init__(self, topic, server, url, endpoints, exchange=None,
                 fanout=False, executor='eventlet'):
        super(DfaRpcServer, self).__init__()
        transport = messaging.get_transport(cfg.CONF, url=url)
        target = messaging.Target(exchange=exchange, topic=topic,
                                  server=server, fanout=fanout)
        endpoints = [endpoints]
        self._server = messaging.get_rpc_server(transport, target, endpoints,
                                                executor=executor)
        LOG.debug('RPC server: topic=%s, server=%s, endpoints=%s' % (
            topic, server, endpoints))

    def start(self):
        if self._server:
            self._server.start()

    def wait(self):
        try:
            while True:
                time.sleep(1)
        except Exception as exc:
            LOG.exception('RPC Server: Exception %s occurred', str(exc))
            self.stop()

    def stop(self):
        if self._server:
            self._server.stop()
            self._server.wait()


class DfaNotificationEndpoints(object):

    """Notification endpoints."""

    def __init__(self, endp):
        self._endpoint = endp

    def info(self, ctxt, publisher_id, event_type, payload, metadata):
        self._endpoint.callback(metadata.get('timestamp'), event_type, payload)


class DfaNotifcationListener(object):

    """RPC Client class for DFA enabler."""

    def __init__(self, topic, url, endpoints, exchange=None, fanout=False):
        super(DfaNotifcationListener, self).__init__()
        transport = messaging.get_transport(cfg.CONF, url=url)
        targets = [messaging.Target(exchange=exchange,
                                    fanout=fanout,
                                    topic=topic)]
        endpoints = [endpoints]
        self._listener = messaging.get_notification_listener(transport,
                                                             targets,
                                                             endpoints)

    def start(self):
        if self._listener:
            self._listener.start()

    def wait(self):
        try:
            while True:
                time.sleep(1)
        except Exception as exc:
            LOG.exception('RPC Server: Exception %s occurred', str(exc))
            self.stop()

    def stop(self):
        if self._listener:
            self._listener.stop()
            self._listener.wait()



# Copyright 2018 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import ssl
from ssl import *  # noqa

from oslo_config import cfg

CONF = cfg.CONF


# The ucsmsdk disables verification of ssl certificates by default. Override
# the SSLContext class to enable certificate checking if https_verify is set to
# True.
class SSLContext(ssl.SSLContext):
    def __init__(self, *args, **kwargs):
        super(SSLContext, self).__init__(*args, **kwargs)
        self.verify_mode = CERT_REQUIRED if (
            CONF.ml2_cisco_ucsm.ucsm_https_verify) else CERT_NONE


# Override wrap_socket function for when the ucsmsdk is using the
# TLS1Connection function
def wrap_socket(sock, keyfile=None, certfile=None,
                server_side=False, cert_reqs=CERT_NONE,
                ssl_version=PROTOCOL_SSLv23, ca_certs=None,
                do_handshake_on_connect=True,
                suppress_ragged_eofs=True,
                ciphers=None):
    cert_reqs = CERT_REQUIRED if (
            CONF.ml2_cisco_ucsm.ucsm_https_verify) else CERT_NONE
    return ssl.wrap_socket(sock=sock, keyfile=keyfile, certfile=certfile,
                     server_side=server_side, cert_reqs=cert_reqs,
                     ssl_version=ssl_version, ca_certs=ca_certs,
                     do_handshake_on_connect=do_handshake_on_connect,
                     suppress_ragged_eofs=suppress_ragged_eofs,
                     ciphers=ciphers)



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Exceptions used by Cisco UCSM ML2 mechanism driver."""

from neutron_lib import exceptions

from networking_cisco._i18n import _


class UcsmConnectFailed(exceptions.NeutronException):
    message = _("Unable to connect to UCS Manager %(ucsm_ip)s. "
                "Reason: %(exc)s.")


class UcsmConfigReadFailed(exceptions.NeutronException):
    message = _("Unable to read config from UCS Manager %(ucsm_ip)s. "
                "Reason: %(exc)s.")


class UcsmConfigFailed(exceptions.NeutronException):
    message = _("Failed to configure %(config)s on UCS Manager %(ucsm_ip)s. "
                "Reason: %(exc)s.")


class UcsmConfigDeleteFailed(exceptions.NeutronException):
    message = _("Failed to delete %(config)s on UCS Manager %(ucsm_ip)s. "
                "Reason: %(exc)s.")


class UcsmDisconnectFailed(exceptions.NeutronException):
    message = _("Disconnect to UCS Manager %(ucsm_ip)s failed. "
                "Reason: %(exc)s.")



# Copyright (c) 2014-2016 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import netaddr
import six

from oslo_config import cfg
from oslo_log import log
from oslo_utils import excutils
import sqlalchemy as sa

from networking_cisco.ml2_drivers.nexus import (
    constants as const)
from networking_cisco.ml2_drivers.nexus import (
    nexus_models_v2)

from networking_cisco._i18n import _
from networking_cisco import backwards_compatibility as bc
from networking_cisco.backwards_compatibility import constants as p_const
from networking_cisco.backwards_compatibility import ml2_api as api

from neutron_lib import exceptions as exc

LOG = log.getLogger(__name__)

# Nexus switches start VNI at 4096 = max VLAN + 2 (2 for reserved VLAN 0, 4095)
MIN_NEXUS_VNI = p_const.MAX_VLAN_TAG + 2

nexus_vxlan_opts = [
    cfg.ListOpt('vni_ranges',
                default=[],
                help=_("Comma-separated list of <vni_min>:<vni_max> tuples "
                       "enumerating ranges of VXLAN Network IDs that are "
                       "available for tenant network allocation. "
                       "Example format: vni_ranges = 100:1000,2000:6000")),
    cfg.ListOpt('mcast_ranges',
                default=[],
                help=_("Multicast groups for the VXLAN interface. When "
                       "configured, will enable sending all broadcast traffic "
                       "to this multicast group. Comma separated list of "
                       "min:max ranges of multicast IP's. "
                       "NOTE: Must be a valid multicast IP, invalid IP's will "
                       "be discarded. "
                       "Example format: mcast_ranges = 224.0.0.1:224.0.0.3, "
                       "224.0.1.1:224.0.1.3"))
]

cfg.CONF.register_opts(nexus_vxlan_opts, "ml2_type_nexus_vxlan")


class NexusVxlanTypeDriver(bc.VXLAN_TUNNEL_TYPE):
    def __init__(self):
        super(NexusVxlanTypeDriver, self).__init__(
            nexus_models_v2.NexusVxlanAllocation)

    def _get_mcast_group_for_vni(self, context, vni):
        session = bc.get_tunnel_session(context)
        mcast_grp = (session.query(nexus_models_v2.NexusMcastGroup).
                     filter_by(associated_vni=vni).first())
        if not mcast_grp:
            mcast_grp = self._allocate_mcast_group(session, vni)
        return mcast_grp

    def get_type(self):
        return const.TYPE_NEXUS_VXLAN

    def initialize(self):
        self.tunnel_ranges = []
        self.conf_mcast_ranges = cfg.CONF.ml2_type_nexus_vxlan.mcast_ranges
        self._verify_vni_ranges()
        self.sync_allocations()

    def _verify_vni_ranges(self):
        try:
            self.conf_vxlan_ranges = self._parse_nexus_vni_ranges(
                cfg.CONF.ml2_type_nexus_vxlan.vni_ranges, self.tunnel_ranges)
            LOG.info("Cisco Nexus VNI ranges: %s", self.conf_vxlan_ranges)
        except Exception:
            LOG.exception("Failed to parse vni_ranges. "
                          "Service terminated!")
            raise SystemExit()

    def _parse_nexus_vni_ranges(self, tunnel_ranges, current_range):
        for entry in tunnel_ranges:
            entry = entry.strip()
            try:
                tun_min, tun_max = entry.split(':')
                tun_min = tun_min.strip()
                tun_max = tun_max.strip()
                tunnel_range = int(tun_min), int(tun_max)
            except ValueError as ex:
                raise exc.NetworkTunnelRangeError(tunnel_range=entry, error=ex)

            self._parse_nexus_vni_range(tunnel_range)
            current_range.append(tunnel_range)

        LOG.info("Nexus VXLAN ID ranges: %(range)s",
                 {'range': current_range})

    def _parse_nexus_vni_range(self, tunnel_range):
        """Raise an exception for invalid tunnel range or malformed range."""
        for ident in tunnel_range:
            if not self._is_valid_nexus_vni(ident):
                raise exc.NetworkTunnelRangeError(
                    tunnel_range=tunnel_range,
                    error=_("%(id)s is not a valid Nexus VNI value.") %
                    {'id': ident})

        if tunnel_range[1] < tunnel_range[0]:
            raise exc.NetworkTunnelRangeError(
                tunnel_range=tunnel_range,
                error=_("End of tunnel range is less than start of "
                        "tunnel range."))

    def _is_valid_nexus_vni(self, vni):
        return MIN_NEXUS_VNI <= vni <= p_const.MAX_VXLAN_VNI

    def _parse_mcast_ranges(self):
        ranges = (range.split(':') for range in self.conf_mcast_ranges)
        for low, high in ranges:
            for mcast_ip in netaddr.iter_iprange(low, high):
                if mcast_ip.is_multicast():
                    yield str(mcast_ip)

    def _allocate_mcast_group(self, session, vni):
        allocs = dict(session.query(nexus_models_v2.
                                    NexusMcastGroup.mcast_group,
                      sa.func.count(nexus_models_v2.
                                    NexusMcastGroup.mcast_group)).
                      group_by(nexus_models_v2.
                               NexusMcastGroup.mcast_group).all())

        mcast_for_vni = None
        for mcast_ip in self._parse_mcast_ranges():
            if not six.u(mcast_ip) in allocs:
                mcast_for_vni = mcast_ip
                break
        try:
            if not mcast_for_vni:
                mcast_for_vni = min(allocs, key=allocs.get)
        except ValueError:
            with excutils.save_and_reraise_exception():
                LOG.exception("Unable to allocate a multicast group for "
                              "VNID:%s", vni)

        alloc = nexus_models_v2.NexusMcastGroup(mcast_group=mcast_for_vni,
                                associated_vni=vni)

        session.add(alloc)
        session.flush()
        return mcast_for_vni

    def allocate_tenant_segment(self, context):
        alloc = self.allocate_partially_specified_segment(context)
        if not alloc:
            return
        vni = alloc.vxlan_vni
        mcast_group = self._get_mcast_group_for_vni(
            context, vni)
        return {api.NETWORK_TYPE: const.TYPE_NEXUS_VXLAN,
                api.PHYSICAL_NETWORK: mcast_group,
                api.SEGMENTATION_ID: alloc.vxlan_vni}

    def sync_allocations(self):
        """
        Synchronize vxlan_allocations table with configured tunnel ranges.
        """

        # determine current configured allocatable vnis
        vxlan_vnis = set()
        for tun_min, tun_max in self.tunnel_ranges:
            vxlan_vnis |= set(six.moves.range(tun_min, tun_max + 1))

        session = bc.get_writer_session()
        with session.begin(subtransactions=True):
            # remove from table unallocated tunnels not currently allocatable
            # fetch results as list via all() because we'll be iterating
            # through them twice
            allocs = (session.query(nexus_models_v2.NexusVxlanAllocation).
                      with_lockmode("update").all())
            # collect all vnis present in db
            existing_vnis = set(alloc.vxlan_vni for alloc in allocs)
            # collect those vnis that needs to be deleted from db
            vnis_to_remove = [alloc.vxlan_vni for alloc in allocs
                              if (alloc.vxlan_vni not in vxlan_vnis and
                                  not alloc.allocated)]
            # Immediately delete vnis in chunks. This leaves no work for
            # flush at the end of transaction
            bulk_size = 100
            chunked_vnis = (vnis_to_remove[i:i + bulk_size] for i in
                            range(0, len(vnis_to_remove), bulk_size))
            for vni_list in chunked_vnis:
                session.query(nexus_models_v2.NexusVxlanAllocation).filter(
                    nexus_models_v2.NexusVxlanAllocation.
                    vxlan_vni.in_(vni_list)).delete(
                        synchronize_session=False)
            # collect vnis that need to be added
            vnis = list(vxlan_vnis - existing_vnis)
            chunked_vnis = (vnis[i:i + bulk_size] for i in
                            range(0, len(vnis), bulk_size))
            for vni_list in chunked_vnis:
                bulk = [{'vxlan_vni': vni, 'allocated': False}
                        for vni in vni_list]
                session.execute(nexus_models_v2.NexusVxlanAllocation.
                                __table__.insert(), bulk)

    def reserve_provider_segment(self, context, segment):
        if self.is_partial_segment(segment):
            alloc = self.allocate_partially_specified_segment(context)
            if not alloc:
                raise exc.NoNetworkAvailable
        else:
            segmentation_id = segment.get(api.SEGMENTATION_ID)
            alloc = self.allocate_fully_specified_segment(
                context, vxlan_vni=segmentation_id)
            if not alloc:
                raise exc.TunnelIdInUse(tunnel_id=segmentation_id)
        return {api.NETWORK_TYPE: p_const.TYPE_VXLAN,
                api.PHYSICAL_NETWORK: None,
                api.SEGMENTATION_ID: alloc.vxlan_vni}

    def release_segment(self, context, segment):
        vxlan_vni = segment[api.SEGMENTATION_ID]

        inside = any(lo <= vxlan_vni <= hi for lo, hi in self.tunnel_ranges)

        session = bc.get_tunnel_session(context)
        with session.begin(subtransactions=True):
            query = (session.query(
                nexus_models_v2.NexusVxlanAllocation).
                     filter_by(vxlan_vni=vxlan_vni))
            if inside:
                count = query.update({"allocated": False})
                if count:
                    mcast_row = (
                         session.query(nexus_models_v2.NexusMcastGroup)
                         .filter_by(associated_vni=vxlan_vni).first())
                    session.delete(mcast_row)
                    LOG.debug("Releasing vxlan tunnel %s to pool",
                              vxlan_vni)
            else:
                count = query.delete()
                if count:
                    LOG.debug("Releasing vxlan tunnel %s outside pool",
                              vxlan_vni)

        if not count:
            LOG.warning("vxlan_vni %s not found", vxlan_vni)

    def add_endpoint(self, ip, udp_port):
        pass

    def delete_endpoint(self, ip):
        pass

    def delete_endpoint_by_host_or_ip(self, host, ip):
        pass

    def get_endpoint_by_host(self, host):
        pass

    def get_endpoint_by_ip(self, ip):
        pass

    def get_endpoints(self):
        pass
